/*******************************************************************************
**	Permedia II chipset driver
********************************************************************************
**
 * Permedia2 framebuffer driver.
 * Copyright (c) 1998-1999 Ilario Nardinocchi (nardinoc@CS.UniBO.IT)
 * Driver used for port under KGI license with permission of Ilario 
 * Nardinocchi.
 * Copyright (c) 1999 Andreas Beck - port to KGI
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	Andreas_Beck
#define	DRIVER_NAME	"Permedia 2 fb-derived chipset driver"
#define	DRIVER_REV	"$Revision: 1.5 $"

#undef DEBUG
#define DEBUG_LEVEL 0

#include <kgi/system.h>
#include <kgi/module.h>

#include <3dlabs/pm2free.h>

#if 0

struct pm2fb_par {
	kgiu32 pixclock;		/* pixclock in KHz */
	kgiu32 width;		/* width of virtual screen */
	kgiu32 height;		/* height of virtual screen */
	kgiu32 hsstart;		/* horiz. sync start */
	kgiu32 hsend;		/* horiz. sync end */
	kgiu32 hbend;		/* horiz. blank end (also gate end) */
	kgiu32 htotal;		/* total width (w/ sync & blank) */
	kgiu32 vsstart;		/* vert. sync start */
	kgiu32 vsend;		/* vert. sync end */
	kgiu32 vbend;		/* vert. blank end */
	kgiu32 vtotal;		/* total height (w/ sync & blank) */
	kgiu32 stride;		/* screen stride */
	kgiu32 base;		/* screen base (xoffset+yoffset) */
	kgiu32 depth;		/* screen depth (8, 16, 24 or 32) */
	kgiu32 video;		/* video control (hsync,vsync) */
};

#define OPTF_OLD_MEM		(1L<<0)
#define OPTF_YPAN		(1L<<1)
#define OPTF_VIRTUAL		(1L<<2)
static struct {
	char font[40];
	kgiu32 flags;
	struct pm2fb_par user_mode;
} pm2fb_options =
	{"\0", 0L, {25174,640,480,4,28,40,199,9,11,45,524,80,0,8,121}};

struct pm2pci_par {
	kgiu32 mem_config;
	kgiu32 mem_control;
	kgiu32 boot_address;
	struct pci_dev* dev;
};

static const char permedia2_name[16]="Permedia2";

static struct pm2fb_info {
/*	struct fb_info_gen gen;		removed for KGI port */
	int board;			/* Permedia2 board index (see
					   board_table[] below) */
	struct {
		unsigned char* fb_base;	/* framebuffer memory base */
		kgiu32  fb_size;		/* framebuffer memory size */
		unsigned char* rg_base;	/* register memory base */
		unsigned char* p_fb;	/* physical address of frame buffer */
		unsigned char* v_fb;	/* virtual address of frame buffer */
		unsigned char* p_regs;	/* physical address of registers
					   region, must be rg_base or
					   rg_base+PM2_REGS_SIZE depending on
					   the host endianness */
		unsigned char* v_regs;	/* virtual address of p_regs */
	} regions;
	union {				/* here, the per-board par structs */
		struct pm2pci_par pci;	/* Permedia2 PCI boards data */
	} board_par;
	struct pm2fb_par current_par;	/* displayed screen */
	int current_par_valid;
	kgiu32 memclock;			/* memclock (set by the per-board
					   		init routine) */
	struct display disp;
	struct {
		kgiu8 transp;
		kgiu8 red;
		kgiu8 green;
		kgiu8 blue;
	} palette[256];
	union {
#ifdef FBCON_HAS_CFB16
		kgiu16 cmap16[16];
#endif
#ifdef FBCON_HAS_CFB24
		kgiu32 cmap24[16];
#endif
#ifdef FBCON_HAS_CFB32
		kgiu32 cmap32[16];
#endif
	} cmap;
} fb_info;

static int pm2pci_detect(struct pm2fb_info*);
static void pm2pci_init(struct pm2fb_info*);

/*
 * Table of the supported Permedia2 based boards.
 * Three hooks are defined for each board:
 * detect(): should return 1 if the related board has been detected, 0
 *           otherwise. It should also fill the fields 'regions.fb_base',
 *           'regions.fb_size', 'regions.rg_base' and 'memclock' in the
 *           passed pm2fb_info structure.
 * init(): called immediately after the reset of the Permedia2 chip.
 *         It should reset the memory controller if needed (the MClk
 *         is set shortly afterwards by the caller).
 * cleanup(): called after the driver has been unregistered.
 *
 * the init and cleanup pointers can be NULL.
 */
static const struct {
	int (*detect)(struct pm2fb_info*);
	void (*init)(struct pm2fb_info*);
	void (*cleanup)(struct pm2fb_info*);
	char name[32];
} board_table[] = {
	{ pm2pci_detect, pm2pci_init, NULL, "Permedia2 PCI board" },
	{ NULL, }
};


static void pm2fb_detect(void);
static int pm2fb_encode_fix(struct fb_fix_screeninfo* fix,
				const void* par, struct fb_info_gen* info);
static int pm2fb_decode_var(const struct fb_var_screeninfo* var,
					void* par, struct fb_info_gen* info);
static int pm2fb_encode_var(struct fb_var_screeninfo* var,
				const void* par, struct fb_info_gen* info);
static void pm2fb_get_par(void* par, struct fb_info_gen* info);
static void pm2fb_set_par(const void* par, struct fb_info_gen* info);
static int pm2fb_getcolreg(unsigned regno,
			unsigned* red, unsigned* green, unsigned* blue,
				unsigned* transp, struct fb_info* info);
static int pm2fb_setcolreg(unsigned regno,
			unsigned red, unsigned green, unsigned blue,
				unsigned transp, struct fb_info* info);
static int pm2fb_blank(int blank_mode, struct fb_info_gen* info);
static int pm2fb_pan_display(const struct fb_var_screeninfo* var,
					struct fb_info_gen* info);
static void pm2fb_set_disp(const void* par, struct display* disp,
						struct fb_info_gen* info);

static struct fbgen_hwswitch pm2fb_hwswitch={
	pm2fb_detect, pm2fb_encode_fix, pm2fb_decode_var,
	pm2fb_encode_var, pm2fb_get_par, pm2fb_set_par,
	pm2fb_getcolreg, pm2fb_setcolreg, pm2fb_pan_display,
	pm2fb_blank, pm2fb_set_disp
};

static int pm2fb_open(struct fb_info* info, int user);
static int pm2fb_release(struct fb_info* info, int user);

static struct fb_ops pm2fb_ops={
	pm2fb_open, pm2fb_release, fbgen_get_fix, fbgen_get_var,
	fbgen_set_var, fbgen_get_cmap, fbgen_set_cmap, fbgen_pan_display,
	NULL /* fb_ioctl() */, NULL /* fb_mmap() */
};

/***************************************************************************
 * Begin of Permedia2 specific functions
 ***************************************************************************/




/*
 * copy with packed pixels (8/16bpp only).
 */
static void pm2fb_pp_copy(struct pm2fb_info* i, kgis32 xsrc, kgis32 ysrc,
					kgis32 x, kgis32 y, kgis32 w, kgis32 h) {
	kgis32 scale=i->current_par.depth==8?2:1;
	kgis32 offset;

	if (!w || !h)
		return;
	WAIT_FIFO(i, 7);
	pm2_WR(i, PM2R_CONFIG,	PM2F_CONFIG_FB_WRITE_ENABLE|
				PM2F_CONFIG_FB_PACKED_DATA|
				PM2F_CONFIG_FB_READ_SOURCE_ENABLE);
	pm2_WR(i, PM2R_FB_PIXEL_OFFSET, 0);
	pm2_WR(i, PM2R_FB_SOURCE_DELTA,	((ysrc-y)&0xfff)<<16|
						((xsrc-x)&0xfff));
	offset=(x&0x3)-(xsrc&0x3);
	pm2_WR(i, PM2R_RECTANGLE_ORIGIN, (y<<16)|(x>>scale));
	pm2_WR(i, PM2R_RECTANGLE_SIZE, (h<<16)|((w+7)>>scale));
	pm2_WR(i, PM2R_PACKED_DATA_LIMITS, (offset<<29)|(x<<16)|(x+w));
	DEFW();
	pm2_WR(i, PM2R_RENDER,	PM2F_RENDER_RECTANGLE|
				(x<xsrc?PM2F_INCREASE_X:0)|
				(y<ysrc?PM2F_INCREASE_Y:0));
	wait_pm2(i);
}

/*
 * block operation. copy=0: rectangle fill, copy=1: rectangle copy.
 */
static void pm2fb_block_op(struct pm2fb_info* i, int copy,
					kgis32 xsrc, kgis32 ysrc,
					kgis32 x, kgis32 y, kgis32 w, kgis32 h,
					kgiu32 color) {

	if (!w || !h)
		return;
	WAIT_FIFO(i, 6);
	pm2_WR(i, PM2R_CONFIG,	PM2F_CONFIG_FB_WRITE_ENABLE|
				PM2F_CONFIG_FB_READ_SOURCE_ENABLE);
	pm2_WR(i, PM2R_FB_PIXEL_OFFSET, 0);
	if (copy)
		pm2_WR(i, PM2R_FB_SOURCE_DELTA,	((ysrc-y)&0xfff)<<16|
							((xsrc-x)&0xfff));
	else
		pm2_WR(i, PM2R_FB_BLOCK_COLOR, color);
	pm2_WR(i, PM2R_RECTANGLE_ORIGIN, (y<<16)|x);
	pm2_WR(i, PM2R_RECTANGLE_SIZE, (h<<16)|w);
	DEFW();
	pm2_WR(i, PM2R_RENDER,	PM2F_RENDER_RECTANGLE|
				(x<xsrc?PM2F_INCREASE_X:0)|
				(y<ysrc?PM2F_INCREASE_Y:0)|
				(copy?0:PM2F_RENDER_FASTFILL));
	wait_pm2(i);
}

/***************************************************************************
 * Begin of generic initialization functions
 ***************************************************************************/


static int pm2fb_conf(struct pm2fb_info* p) {

	for (p->board=0; board_table[p->board].detect &&
			!(board_table[p->board].detect(p)); p->board++);
	if (!board_table[p->board].detect) {
		DPRINTK("no board found.\n");
		return 0;
	}
	DPRINTK("found board: %s\n", board_table[p->board].name);
	p->regions.p_fb=p->regions.fb_base;
	p->regions.v_fb=MMAP(p->regions.p_fb, p->regions.fb_size);
#ifdef __LITTLE_ENDIAN
	p->regions.p_regs=p->regions.rg_base;
#else
	p->regions.p_regs=p->regions.rg_base+PM2_REGS_SIZE;
#endif
	p->regions.v_regs=MMAP(p->regions.p_regs, PM2_REGS_SIZE);
	return 1;
}

/***************************************************************************
 * Begin of per-board initialization functions
 ***************************************************************************/


static void pm2pci_init(struct pm2fb_info* p) {
	struct pm2pci_par* pci=&p->board_par.pci;

	WAIT_FIFO(p, 3);
	pm2_WR(p, PM2R_MEM_CONTROL, pci->mem_control);
	pm2_WR(p, PM2R_BOOT_ADDRESS, pci->boot_address);
	DEFW();
	pm2_WR(p, PM2R_MEM_CONFIG, pci->mem_config);
}

/***************************************************************************
 * Console hw acceleration
 ***************************************************************************/


static int pm2fb_blank(int blank_mode, struct fb_info_gen* info) {
	struct pm2fb_info* i=(struct pm2fb_info* )info;
	kgiu32 video;

	if (!i->current_par_valid)
		return 1;
	video=i->current_par.video;
	if (blank_mode>0) {
		switch (blank_mode-1) {
			case VESA_NO_BLANKING:		/* FIXME */
				video=video&~(PM2F_VIDEO_ENABLE);
				break;
			case VESA_HSYNC_SUSPEND:
				video=video&~(PM2F_HSYNC_MASK|
						PM2F_BLANK_LOW);
				break;
			case VESA_VSYNC_SUSPEND:
				video=video&~(PM2F_VSYNC_MASK|
						PM2F_BLANK_LOW);
				break;
			case VESA_POWERDOWN:
				video=video&~(PM2F_VSYNC_MASK|
						PM2F_HSYNC_MASK|
						PM2F_BLANK_LOW);
				break;
		}
	}
	WAIT_FIFO(i, 1);
	pm2_WR(i, PM2R_VIDEO_CONTROL, video);
	return 0;
}


static void pm2fb_pp_bmove(struct display* p, int sy, int sx,
				int dy, int dx, int height, int width) {

	if (fontwidthlog(p)) {
		sx=sx<<fontwidthlog(p);
		dx=dx<<fontwidthlog(p);
		width=width<<fontwidthlog(p);
	}
	else {
		sx=sx*fontwidth(p);
		dx=dx*fontwidth(p);
		width=width*fontwidth(p);
	}
	sy=sy*fontheight(p);
	dy=dy*fontheight(p);
	height=height*fontheight(p);
	pm2fb_pp_copy((struct pm2fb_info* )p->fb_info, sx, sy, dx,
							dy, width, height);
}

static void pm2fb_bmove(struct display* p, int sy, int sx,
				int dy, int dx, int height, int width) {

	if (fontwidthlog(p)) {
		sx=sx<<fontwidthlog(p);
		dx=dx<<fontwidthlog(p);
		width=width<<fontwidthlog(p);
	}
	else {
		sx=sx*fontwidth(p);
		dx=dx*fontwidth(p);
		width=width*fontwidth(p);
	}
	sy=sy*fontheight(p);
	dy=dy*fontheight(p);
	height=height*fontheight(p);
	pm2fb_block_op((struct pm2fb_info* )p->fb_info, 1, sx, sy, dx, dy,
							width, height, 0);
}

#ifdef FBCON_HAS_CFB8
static void pm2fb_clear8(struct vc_data* conp, struct display* p,
				int sy, int sx, int height, int width) {
	kgiu32 c;

	sx=sx*fontwidth(p);
	width=width*fontwidth(p);
	sy=sy*fontheight(p);
	height=height*fontheight(p);
	c=attr_bgcol_ec(p, conp);
	c|=c<<8;
	c|=c<<16;
	pm2fb_block_op((struct pm2fb_info* )p->fb_info, 0, 0, 0, sx, sy,
							width, height, c);
}

static void pm2fb_clear_margins8(struct vc_data* conp, struct display* p,
							int bottom_only) {
	kgiu32 c;
	kgiu32 sx;
	kgiu32 sy;

	c=attr_bgcol_ec(p, conp);
	c|=c<<8;
	c|=c<<16;
	sx=conp->vc_cols*fontwidth(p);
	sy=conp->vc_rows*fontheight(p);
	if (!bottom_only)
		pm2fb_block_op((struct pm2fb_info* )p->fb_info, 0, 0, 0,
			sx, 0, (p->var.xres-sx), p->var.yres_virtual, c);
	pm2fb_block_op((struct pm2fb_info* )p->fb_info, 0, 0, 0,
				0, p->var.yoffset+sy, sx, p->var.yres-sy, c);
}

static struct display_switch pm2_cfb8 = {
	fbcon_cfb8_setup, pm2fb_pp_bmove, pm2fb_clear8,
	fbcon_cfb8_putc, fbcon_cfb8_putcs, fbcon_cfb8_revc,
	NULL /* cursor() */, NULL /* set_font() */,
	pm2fb_clear_margins8,
	FONTWIDTH(4)|FONTWIDTH(8)|FONTWIDTH(12)|FONTWIDTH(16) };
#endif /* FBCON_HAS_CFB8 */

#ifdef FBCON_HAS_CFB16
static void pm2fb_clear16(struct vc_data* conp, struct display* p,
				int sy, int sx, int height, int width) {
	kgiu32 c;

	sx=sx*fontwidth(p);
	width=width*fontwidth(p);
	sy=sy*fontheight(p);
	height=height*fontheight(p);
	c=((kgiu16 *)p->dispsw_data)[attr_bgcol_ec(p, conp)];
	c|=c<<16;
	pm2fb_block_op((struct pm2fb_info* )p->fb_info, 0, 0, 0, sx, sy,
							width, height, c);
}

static void pm2fb_clear_margins16(struct vc_data* conp, struct display* p,
							int bottom_only) {
	kgiu32 c;
	kgiu32 sx;
	kgiu32 sy;

	c = ((kgiu16 *)p->dispsw_data)[attr_bgcol_ec(p, conp)];
	c|=c<<16;
	sx=conp->vc_cols*fontwidth(p);
	sy=conp->vc_rows*fontheight(p);
	if (!bottom_only)
		pm2fb_block_op((struct pm2fb_info* )p->fb_info, 0, 0, 0,
			sx, 0, (p->var.xres-sx), p->var.yres_virtual, c);
	pm2fb_block_op((struct pm2fb_info* )p->fb_info, 0, 0, 0,
				0, p->var.yoffset+sy, sx, p->var.yres-sy, c);
}

static struct display_switch pm2_cfb16 = {
	fbcon_cfb16_setup, pm2fb_pp_bmove, pm2fb_clear16,
	fbcon_cfb16_putc, fbcon_cfb16_putcs, fbcon_cfb16_revc,
	NULL /* cursor() */, NULL /* set_font() */,
	pm2fb_clear_margins16,
	FONTWIDTH(4)|FONTWIDTH(8)|FONTWIDTH(12)|FONTWIDTH(16) };
#endif /* FBCON_HAS_CFB16 */

#ifdef FBCON_HAS_CFB24
/*
 * fast fill for 24bpp works only when red==green==blue
 */
static void pm2fb_clear24(struct vc_data* conp, struct display* p,
				int sy, int sx, int height, int width) {
	struct pm2fb_info* i=(struct pm2fb_info* )p->fb_info;
	kgiu32 c;

	c=attr_bgcol_ec(p, conp);
	if (		i->palette[c].red==i->palette[c].green &&
			i->palette[c].green==i->palette[c].blue) {
		c=((kgiu32 *)p->dispsw_data)[c];
		c|=(c&0xff0000)<<8;
		sx=sx*fontwidth(p);
		width=width*fontwidth(p);
		sy=sy*fontheight(p);
		height=height*fontheight(p);
		pm2fb_block_op(i, 0, 0, 0, sx, sy, width, height, c);
	}
	else
		fbcon_cfb24_clear(conp, p, sy, sx, height, width);

}

static void pm2fb_clear_margins24(struct vc_data* conp, struct display* p,
							int bottom_only) {
	struct pm2fb_info* i=(struct pm2fb_info* )p->fb_info;
	kgiu32 c;
	kgiu32 sx;
	kgiu32 sy;

	c=attr_bgcol_ec(p, conp);
	if (		i->palette[c].red==i->palette[c].green &&
			i->palette[c].green==i->palette[c].blue) {
		c=((kgiu32 *)p->dispsw_data)[c];
		c|=(c&0xff0000)<<8;
		sx=conp->vc_cols*fontwidth(p);
		sy=conp->vc_rows*fontheight(p);
		if (!bottom_only)
		pm2fb_block_op(i, 0, 0, 0, sx, 0, (p->var.xres-sx),
							p->var.yres_virtual, c);
		pm2fb_block_op(i, 0, 0, 0, 0, p->var.yoffset+sy,
						sx, p->var.yres-sy, c);
	}
	else
		fbcon_cfb24_clear_margins(conp, p, bottom_only);

}

static struct display_switch pm2_cfb24 = {
	fbcon_cfb24_setup, pm2fb_bmove, pm2fb_clear24,
	fbcon_cfb24_putc, fbcon_cfb24_putcs, fbcon_cfb24_revc,
	NULL /* cursor() */, NULL /* set_font() */,
	pm2fb_clear_margins24,
	FONTWIDTH(4)|FONTWIDTH(8)|FONTWIDTH(12)|FONTWIDTH(16) };
#endif /* FBCON_HAS_CFB24 */

#ifdef FBCON_HAS_CFB32
static void pm2fb_clear32(struct vc_data* conp, struct display* p,
				int sy, int sx, int height, int width) {
	kgiu32 c;

	sx=sx*fontwidth(p);
	width=width*fontwidth(p);
	sy=sy*fontheight(p);
	height=height*fontheight(p);
	c=((kgiu32 *)p->dispsw_data)[attr_bgcol_ec(p, conp)];
	pm2fb_block_op((struct pm2fb_info* )p->fb_info, 0, 0, 0, sx, sy,
							width, height, c);
}

static void pm2fb_clear_margins32(struct vc_data* conp, struct display* p,
							int bottom_only) {
	kgiu32 c;
	kgiu32 sx;
	kgiu32 sy;

	c = ((kgiu32 *)p->dispsw_data)[attr_bgcol_ec(p, conp)];
	sx=conp->vc_cols*fontwidth(p);
	sy=conp->vc_rows*fontheight(p);
	if (!bottom_only)
		pm2fb_block_op((struct pm2fb_info* )p->fb_info, 0, 0, 0,
			sx, 0, (p->var.xres-sx), p->var.yres_virtual, c);
	pm2fb_block_op((struct pm2fb_info* )p->fb_info, 0, 0, 0,
				0, p->var.yoffset+sy, sx, p->var.yres-sy, c);
}

static struct display_switch pm2_cfb32 = {
	fbcon_cfb32_setup, pm2fb_bmove, pm2fb_clear32,
	fbcon_cfb32_putc, fbcon_cfb32_putcs, fbcon_cfb32_revc,
	NULL /* cursor() */, NULL /* set_font() */,
	pm2fb_clear_margins32,
	FONTWIDTH(4)|FONTWIDTH(8)|FONTWIDTH(12)|FONTWIDTH(16) };
#endif /* FBCON_HAS_CFB32 */

/***************************************************************************
 * Framebuffer functions
 ***************************************************************************/

static int pm2fb_encode_fix(struct fb_fix_screeninfo* fix,
			const void* par, struct fb_info_gen* info) {
	struct pm2fb_info* i=(struct pm2fb_info* )info;
	struct pm2fb_par* p=(struct pm2fb_par* )par;

	strcpy(fix->id, permedia2_name);
	fix->smem_start=i->regions.p_fb;
	fix->smem_len=i->regions.fb_size;
	fix->mmio_start=i->regions.p_regs;
	fix->mmio_len=PM2_REGS_SIZE;
	fix->accel=FB_ACCEL_3DLABS_PERMEDIA2;
	fix->type=FB_TYPE_PACKED_PIXELS;
	fix->visual=p->depth==8?FB_VISUAL_PSEUDOCOLOR:FB_VISUAL_TRUECOLOR;
	fix->line_length=0;
	fix->xpanstep=p->depth==24?8:64/p->depth;
	fix->ypanstep=1;
	fix->ywrapstep=0;
	return 0;
}

#ifdef PM2FB_MASTER_DEBUG
static void pm2fb_display_var(const struct fb_var_screeninfo* var) {

	kgi_printk( KERN_DEBUG
"- struct fb_var_screeninfo ---------------------------------------------------\n");
	kgi_printk( KERN_DEBUG
		"resolution: %ux%ux%u (virtual %ux%u+%u+%u)\n",
			var->xres, var->yres, var->bits_per_pixel,
			var->xres_virtual, var->yres_virtual,
			var->xoffset, var->yoffset);
	kgi_printk( KERN_DEBUG
		"color: %c%c "
		"R(%u,%u,%u), G(%u,%u,%u), B(%u,%u,%u), T(%u,%u,%u)\n",
			var->grayscale?'G':'C', var->nonstd?'N':'S',
			var->red.offset, var->red.length, var->red.msb_right,
			var->green.offset, var->green.length, var->green.msb_right,
			var->blue.offset, var->blue.length, var->blue.msb_right,
			var->transp.offset, var->transp.length,
			var->transp.msb_right);
	kgi_printk( KERN_DEBUG
		"timings: %ups (%u,%u)-(%u,%u)+%u+%u\n",
		var->pixclock,
		var->left_margin, var->upper_margin, var->right_margin,
		var->lower_margin, var->hsync_len, var->vsync_len);
	kgi_printk(	KERN_DEBUG
		"activate %08x accel_flags %08x sync %08x vmode %08x\n",
		var->activate, var->accel_flags, var->sync, var->vmode);
	kgi_printk(	KERN_DEBUG
"------------------------------------------------------------------------------\n");
}

#define pm2fb_decode_var pm2fb_wrapped_decode_var
#endif


#ifdef PM2FB_MASTER_DEBUG
#undef pm2fb_decode_var

static int pm2fb_decode_var(const struct fb_var_screeninfo* var,
				void* par, struct fb_info_gen* info) {
	int result;

	result=pm2fb_wrapped_decode_var(var, par, info);
	pm2fb_display_var(var);
	return result;
}
#endif

static int pm2fb_encode_var(struct fb_var_screeninfo* var,
				const void* par, struct fb_info_gen* info) {
	struct pm2fb_par* p=(struct pm2fb_par* )par;
	struct fb_var_screeninfo v;
	kgiu32 base;

	memset(&v, 0, sizeof(struct fb_var_screeninfo));
	v.xres_virtual=p->width;
	v.yres_virtual=p->height;
	v.xres=(p->htotal+1)-p->hbend;
	v.yres=(p->vtotal+1)-p->vbend;
	v.right_margin=p->hsstart;
	v.hsync_len=p->hsend-p->hsstart;
	v.left_margin=p->hbend-p->hsend;
	v.lower_margin=p->vsstart+1;
	v.vsync_len=p->vsend-v.lower_margin+1;
	v.upper_margin=p->vbend-v.lower_margin-v.vsync_len;
	v.bits_per_pixel=p->depth;
	if (p->video & PM2F_DATA_64_ENABLE) {
		v.xres=v.xres<<1;
		v.right_margin=v.right_margin<<1;
		v.hsync_len=v.hsync_len<<1;
		v.left_margin=v.left_margin<<1;
	}
	switch (p->depth) {
		case 8:
			v.red.length=v.green.length=v.blue.length=8;
			v.xres=v.xres<<2;
			v.right_margin=v.right_margin<<2;
			v.hsync_len=v.hsync_len<<2;
			v.left_margin=v.left_margin<<2;
			break;
		case 16:
			v.red.offset=11;
			v.red.length=5;
			v.green.offset=5;
			v.green.length=6;
			v.blue.length=5;
			v.xres=v.xres<<1;
			v.right_margin=v.right_margin<<1;
			v.hsync_len=v.hsync_len<<1;
			v.left_margin=v.left_margin<<1;
			break;
		case 32:
			v.transp.offset=24;
			v.red.offset=16;
			v.green.offset=8;
			v.red.length=v.green.length=v.blue.length=
							v.transp.length=8;
			break;
		case 24:
			v.blue.offset=16;
			v.green.offset=8;
			v.red.length=v.green.length=v.blue.length=8;
			v.xres=(v.xres<<2)/3;
			v.right_margin=(v.right_margin<<2)/3;
			v.hsync_len=(v.hsync_len<<2)/3;
			v.left_margin=(v.left_margin<<2)/3;
			break;
	}
	base=from3264(p->base, p->depth, 1);
	v.xoffset=base%v.xres;
	v.yoffset=base/v.xres;
	v.height=v.width=-1;
	v.pixclock=KHZ2PICOS(p->pixclock);
	if ((p->video & PM2F_HSYNC_MASK)==PM2F_HSYNC_ACT_HIGH)
		v.sync|=FB_SYNC_HOR_HIGH_ACT;
	if ((p->video & PM2F_VSYNC_MASK)==PM2F_VSYNC_ACT_HIGH)
		v.sync|=FB_SYNC_VERT_HIGH_ACT;
	if (p->video & PM2F_LINE_DOUBLE)
		v.vmode=FB_VMODE_DOUBLE;
	*var=v;
	return 0;
}

static void set_user_mode(struct pm2fb_info* i) {

	memcpy(&i->current_par, &pm2fb_options.user_mode,
						sizeof(i->current_par));
	if (pm2fb_options.flags & OPTF_YPAN) {
		i->current_par.height=i->regions.fb_size/
			(i->current_par.width*i->current_par.depth/8);
		i->current_par.height=MIN(i->current_par.height,2047);
		i->current_par.height=MAX(i->current_par.height,
						pm2fb_options.user_mode.height);
	}
}

static void pm2fb_get_par(void* par, struct fb_info_gen* info) {
	struct pm2fb_info* i=(struct pm2fb_info* )info;
	
	if (!i->current_par_valid) {
		set_user_mode(i);
		pm2fb_reset(i);
		set_screen(i, &i->current_par);
		i->current_par_valid=1;
	}
	*((struct pm2fb_par* )par)=i->current_par;
}

static void pm2fb_set_par(const void* par, struct fb_info_gen* info) {
	struct pm2fb_info* i=(struct pm2fb_info* )info;
	struct pm2fb_par* p;

	p=(struct pm2fb_par* )par;
	if (i->current_par_valid) {
		i->current_par.base=p->base;
		if (!memcmp(p, &i->current_par, sizeof(struct pm2fb_par))) {
			WAIT_FIFO(i, 1);
			pm2_WR(i, PM2R_SCREEN_BASE, p->base);
			return;
		}
	}
	i->current_par=*p;
	i->current_par_valid=1;
	set_screen(i, &i->current_par);
}

static int pm2fb_getcolreg(unsigned regno,
			unsigned* red, unsigned* green, unsigned* blue,
				unsigned* transp, struct fb_info* info) {
	struct pm2fb_info* i=(struct pm2fb_info* )info;

	if (regno<256) {
		*red=i->palette[regno].red<<8|i->palette[regno].red;
		*green=i->palette[regno].green<<8|i->palette[regno].green;
		*blue=i->palette[regno].blue<<8|i->palette[regno].blue;
		*transp=i->palette[regno].transp<<8|i->palette[regno].transp;
	}
	return regno>255;
}

static int pm2fb_setcolreg(unsigned regno,
			unsigned red, unsigned green, unsigned blue,
				unsigned transp, struct fb_info* info) {
	struct pm2fb_info* i=(struct pm2fb_info* )info;

	if (regno<16) {
		switch (i->current_par.depth) {
#ifdef FBCON_HAS_CFB8
			case 8:
				break;
#endif
#ifdef FBCON_HAS_CFB16
			case 16:
				i->cmap.cmap16[regno]=
					((kgiu32 )red & 0xf800) |
					(((kgiu32 )green & 0xfc00)>>5) |
					(((kgiu32 )blue & 0xf800)>>11);
				break;
#endif
#ifdef FBCON_HAS_CFB24
			case 24:
				i->cmap.cmap24[regno]=
					(((kgiu32 )blue & 0xff00) << 8) |
					((kgiu32 )green & 0xff00) |
					(((kgiu32 )red & 0xff00) >> 8);
				break;
#endif
#ifdef FBCON_HAS_CFB32
			case 32:
	   			i->cmap.cmap32[regno]=
					(((kgiu32 )transp & 0xff00) << 16) |
		    			(((kgiu32 )red & 0xff00) << 8) |
					(((kgiu32 )green & 0xff00)) |
			 		(((kgiu32 )blue & 0xff00) >> 8);
				break;
#endif
			default:
				DPRINTK("bad depth %u\n",
						i->current_par.depth);
				break;
		}
	}
	if (regno<256) {
		i->palette[regno].red=red >> 8;
		i->palette[regno].green=green >> 8;
		i->palette[regno].blue=blue >> 8;
		i->palette[regno].transp=transp >> 8;
		if (i->current_par.depth==8)
			set_color(i, regno, red>>8, green>>8, blue>>8);
	}
	return regno>255;
}

static void pm2fb_set_disp(const void* par, struct display* disp,
						struct fb_info_gen* info) {
	struct pm2fb_info* i=(struct pm2fb_info* )info;
	kgiu32 flags;
	kgiu32 depth;

	save_flags(flags);
	cli();
	disp->screen_base=i->regions.v_fb;
	switch (depth=((struct pm2fb_par* )par)->depth) {
#ifdef FBCON_HAS_CFB8
		case 8:
			disp->dispsw=&pm2_cfb8;
			break;
#endif
#ifdef FBCON_HAS_CFB16
		case 16:
			disp->dispsw=&pm2_cfb16;
			disp->dispsw_data=i->cmap.cmap16;
			break;
#endif
#ifdef FBCON_HAS_CFB24
		case 24:
			disp->dispsw=&pm2_cfb24;
			disp->dispsw_data=i->cmap.cmap24;
			break;
#endif
#ifdef FBCON_HAS_CFB32
		case 32:
			disp->dispsw=&pm2_cfb32;
			disp->dispsw_data=i->cmap.cmap32;
			break;
#endif
		default:
			disp->dispsw=&fbcon_dummy;
			break;
	}
	restore_flags(flags);
}

/***************************************************************************
 * Begin of public functions
 ***************************************************************************/

void pm2fb_mode_setup(char* options) {
	int i;

	for (i=0; user_mode[i].name[0] &&
		strcmp(options, user_mode[i].name); i++);
	if (user_mode[i].name[0])
		memcpy(&pm2fb_options.user_mode, &user_mode[i].par,
					sizeof(pm2fb_options.user_mode));
}

void pm2fb_font_setup(char* options) {

	strncpy(pm2fb_options.font, options, sizeof(pm2fb_options.font));
	pm2fb_options.font[sizeof(pm2fb_options.font)-1]='\0';
}

void pm2fb_setup(char* options, int* ints) {
	char* next;

	while (options) {
		if ((next=strchr(options, ',')))
			*(next++)='\0';
		if (!strncmp(options, "font:", 5))
			pm2fb_font_setup(options+5);
		else if (!strncmp(options, "mode:", 5))
			pm2fb_mode_setup(options+5);
		else if (!strcmp(options, "ypan"))
			pm2fb_options.flags |= OPTF_YPAN;
		else if (!strcmp(options, "oldmem"))
			pm2fb_options.flags |= OPTF_OLD_MEM;
		else if (!strcmp(options, "virtual"))
			pm2fb_options.flags |= OPTF_VIRTUAL;
		options=next;
	}
}

#endif

static const struct kgi_chipset pm2free_info =
{
	"3dlabs",			/* manufacturer		*/
	"Permedia 2",			/* model (detected)	*/
	GM_8BIT | 			/* supported modes	*/
	GM_16BIT| GM_24BIT |
	GM_32BIT,
	0,
	{ 2048, 2048 },			/* maximum resolution	*/
	NULL,				/* Private data hook	*/
	0,				/* RAM size (detected)	*/
	{ 10000000, 230000000 }		/* DCLK limits		*/
};

struct statetable {};

/*	This table holds the default bits for text & graphics mode and
**	a copy of it will be filled in completely during mode_init().
*/
static struct statetable default_state =
{
};

static struct statetable reset_state, debug_state;

/*
**	ressources used by the Permedia 2
*/
static const struct mem_region pm2free_mem_regs =
{
	MEM_NULL,
	0x00000,
	0x10000,
	-1,
	"Permedia 2 registers"
};

static const struct mem_region pm2free_mem_fb =
{
	MEM_NULL,
	0,
	0x00800000,
	-1,
	"Permedia 2 LAW"
};

#ifdef CONFIG_FB_PM2_FIFO_DISCONNECT
#define WAIT_FIFO(p,a)
#else
inline static void WAIT_FIFO(struct chipset_private* p, kgiu32 a) {
	while(pm2_RD(p, PM2R_IN_FIFO_SPACE)<a);
	DEFRW();
}
#endif

static void mnp(kgiu32 clk, unsigned char* mm, unsigned char* nn,
							unsigned char* pp) {
	unsigned char m;
	unsigned char n;
	unsigned char p;
	kgiu32 f;
	kgis32 curr;
	kgis32 delta=100000;

	*mm=*nn=*pp=0;
	for (n=2; n<15; n++) {
		for (m=2; m; m++) {
			f=PM2_REFERENCE_CLOCK*m/n;
			if (f>=150000 && f<=300000) {
				for (p=0; p<5; p++, f>>=1) {
					curr=clk>f?clk-f:f-clk;
					if (curr<delta) {
						delta=curr;
						*mm=m;
						*nn=n;
						*pp=p;
					}
				}
			}
		}
	}
}
static void set_memclock(struct chipset_private *private, kgiu32 clk) {
	int i;
	unsigned char m, n, p;

	mnp(clk, &m, &n, &p);
	WAIT_FIFO(private, 5);
	pm2_RDAC_WR(private, PM2I_RD_MEMORY_CLOCK_3, 6);
	DEFW();
	pm2_RDAC_WR(private, PM2I_RD_MEMORY_CLOCK_1, m);
	pm2_RDAC_WR(private, PM2I_RD_MEMORY_CLOCK_2, n);
	DEFW();
	pm2_RDAC_WR(private, PM2I_RD_MEMORY_CLOCK_3, 8|p);
	DEFW();
	pm2_RDAC_RD(private, PM2I_RD_MEMORY_CLOCK_STATUS);
	DEFR();
	for (i=256; i &&
		!(pm2_RD(private, PM2R_RD_INDEXED_DATA)&PM2F_PLL_LOCKED); i--);
}

#if 0
/* called from DAC driver
 */
void PM2_set_color(struct chipset_private* p, unsigned char regno,
			unsigned char r, unsigned char g, unsigned char b) {

	WAIT_FIFO(p, 4);
	pm2_WR(p, PM2R_RD_PALETTE_WRITE_ADDRESS, regno);
	DEFW();
	pm2_WR(p, PM2R_RD_PALETTE_DATA, r);
	DEFW();
	pm2_WR(p, PM2R_RD_PALETTE_DATA, g);
	DEFW();
	pm2_WR(p, PM2R_RD_PALETTE_DATA, b);
}
#endif

static void set_aperture(struct chipset_private* i, kgi_graphtype mode) {

	WAIT_FIFO(i, 2);
#ifdef __LITTLE_ENDIAN
	pm2_WR(i, PM2R_APERTURE_ONE, 0);	/* FIXME */
	pm2_WR(i, PM2R_APERTURE_TWO, 0);
#else
	switch (mode) {
		case KGIGT_8BIT:
		case KGIGT_24BIT:
			pm2_WR(i, PM2R_APERTURE_ONE, 0);
			pm2_WR(i, PM2R_APERTURE_TWO, 1);
			break;
		case KGIGT_16BIT:
			pm2_WR(i, PM2R_APERTURE_ONE, 2);
			pm2_WR(i, PM2R_APERTURE_TWO, 1);
			break;
		case KGIGT_32BIT:
			pm2_WR(i, PM2R_APERTURE_ONE, 1);
			pm2_WR(i, PM2R_APERTURE_TWO, 1);
			break;
	}
#endif
}

/*
 * partial products for the supported horizontal resolutions.
 */
#define PACKPP(p0,p1,p2)	(((p2)<<6)|((p1)<<3)|(p0))
static const struct {
	kgiu16 width;
	kgiu16 pp;
} pp_table[] = {
	{ 32,	PACKPP(1, 0, 0) }, { 64,	PACKPP(1, 1, 0) },
	{ 96,	PACKPP(1, 1, 1) }, { 128,	PACKPP(2, 1, 1) },
	{ 160,	PACKPP(2, 2, 1) }, { 192,	PACKPP(2, 2, 2) },
	{ 224,	PACKPP(3, 2, 1) }, { 256,	PACKPP(3, 2, 2) },
	{ 288,	PACKPP(3, 3, 1) }, { 320,	PACKPP(3, 3, 2) },
	{ 384,	PACKPP(3, 3, 3) }, { 416,	PACKPP(4, 3, 1) },
	{ 448,	PACKPP(4, 3, 2) }, { 512,	PACKPP(4, 3, 3) },
	{ 544,	PACKPP(4, 4, 1) }, { 576,	PACKPP(4, 4, 2) },
	{ 640,	PACKPP(4, 4, 3) }, { 768,	PACKPP(4, 4, 4) },
	{ 800,	PACKPP(5, 4, 1) }, { 832,	PACKPP(5, 4, 2) },
	{ 896,	PACKPP(5, 4, 3) }, { 1024,	PACKPP(5, 4, 4) },
	{ 1056,	PACKPP(5, 5, 1) }, { 1088,	PACKPP(5, 5, 2) },
	{ 1152,	PACKPP(5, 5, 3) }, { 1280,	PACKPP(5, 5, 4) },
	{ 1536,	PACKPP(5, 5, 5) }, { 1568,	PACKPP(6, 5, 1) },
	{ 1600,	PACKPP(6, 5, 2) }, { 1664,	PACKPP(6, 5, 3) },
	{ 1792,	PACKPP(6, 5, 4) }, { 2048,	PACKPP(6, 5, 5) },
	{ 0,	0 } };

static kgiu32 partprod(kgiu32 xres) {
	int i;

	for (i=0; pp_table[i].width && pp_table[i].width!=xres; i++);
	if (!pp_table[i].width)
		DPRINTK("invalid width %u\n", xres);
	return pp_table[i].pp;
}

static kgiu32 to3264(kgiu32 timing, int bpp, int is64) {

	switch (bpp) {
		case 8:
			timing=timing>>(2+is64);
			break;
		case 16:
			timing=timing>>(1+is64);
			break;
		case 24:
			timing=(timing*3)>>(2+is64);
			break;
		case 32:
			if (is64)
				timing=timing>>1;
			break;
	}
	return timing;
}

static kgiu32 from3264(kgiu32 timing, int bpp, int is64) {

	switch (bpp) {
		case 8:
			timing=timing<<(2+is64);
			break;
		case 16:
			timing=timing<<(1+is64);
			break;
		case 24:
			timing=(timing<<(2+is64))/3;
			break;
		case 32:
			if (is64)
				timing=timing<<1;
			break;
	}
	return timing;
}


static void wait_pm2(struct chipset_private* i) {

	WAIT_FIFO(i, 1);
	pm2_WR(i, PM2R_SYNC, 0);
	DEFRW();
	do {
		while (pm2_RD(i, PM2R_OUT_FIFO_WORDS)==0);
		DEFR();
	} while (pm2_RD(i, PM2R_OUT_FIFO)!=PM2TAG(PM2R_SYNC));
}


/* FIXME - move to clock driver.
 */
static void set_pixclock(struct chipset_private* info, kgiu32 clk) {
	int i;
	unsigned char m, n, p;

	mnp(clk, &m, &n, &p);
	WAIT_FIFO(info, 5);
	pm2_RDAC_WR(info, PM2I_RD_PIXEL_CLOCK_A3, 0);
	DEFW();
	pm2_RDAC_WR(info, PM2I_RD_PIXEL_CLOCK_A1, m);
	pm2_RDAC_WR(info, PM2I_RD_PIXEL_CLOCK_A2, n);
	DEFW();
	pm2_RDAC_WR(info, PM2I_RD_PIXEL_CLOCK_A3, 8|p);
	DEFW();
	pm2_RDAC_RD(info, PM2I_RD_PIXEL_CLOCK_STATUS);
	DEFR();
	for (i=256; i &&
		!(pm2_RD(info, PM2R_RD_INDEXED_DATA)&PM2F_PLL_LOCKED); i--);
}

static int get_bpp(struct kgi_mode *mode) {
	switch (mode->request.graphtype) {
		case KGIGT_8BIT:  return  8;
		case KGIGT_16BIT: return 16;
		case KGIGT_32BIT: return 32;
		case KGIGT_24BIT: return 24;
		default:
			return 0;	/* WHAAAT ? We should never have gotten here ... */
	}
}

static int pm2fb_pan_display(struct chipset_private* private,
					kgiu32 base) {
	WAIT_FIFO(private, 1);
	pm2_WR(private, PM2R_SCREEN_BASE, base);
	return 0;
}

void pm2free_set_origin (struct kgi_display *dpy, kgi_uint x, uint y)
{
	GET_PRIVATE;
	kgiu32 base=to3264(y*dpy->mode.request.virt.x+x, get_bpp(&dpy->mode), 1);
	pm2fb_pan_display(private,base);
}

void pm2free_set_frame (struct kgi_display *dpy, int frame)
{
	GET_PRIVATE;
	kgiu32 base=to3264(dpy->mode.request.virt.y*dpy->mode.request.virt.x*
				frame, get_bpp(&dpy->mode),1);
	pm2fb_pan_display(private,base);
}

static int pm2free_reset(struct kgi_display *dpy)
{
	GET_PRIVATE;

	pm2_WR(private, PM2R_RESET_STATUS, 0);
	DEFRW();
	while (pm2_RD(private, PM2R_RESET_STATUS)&PM2F_BEING_RESET);
	DEFRW();
#ifdef CONFIG_FB_PM2_FIFO_DISCONNECT
	DPRINTK("FIFO disconnect enabled\n");
	pm2_WR(private, PM2R_FIFO_DISCON, 1);
	DEFRW();
#endif
#if 0
	if (board_table[private->board].init)
		board_table[private->board].init(private);
#endif
	WAIT_FIFO(private, 48);
	pm2_WR(private, PM2R_CHIP_CONFIG, pm2_RD(private, PM2R_CHIP_CONFIG)&
					~(PM2F_VGA_ENABLE|PM2F_VGA_FIXED));
	pm2_WR(private, PM2R_BYPASS_WRITE_MASK, ~(0L));
	pm2_WR(private, PM2R_FRAMEBUFFER_WRITE_MASK, ~(0L));
	pm2_WR(private, PM2R_FIFO_CONTROL, 0);
	pm2_WR(private, PM2R_FILTER_MODE, PM2F_SYNCHRONIZATION);
	pm2_WR(private, PM2R_APERTURE_ONE, 0);
	pm2_WR(private, PM2R_APERTURE_TWO, 0);
	pm2_WR(private, PM2R_LB_READ_FORMAT, 0);
	pm2_WR(private, PM2R_LB_WRITE_FORMAT, 0); 
	pm2_WR(private, PM2R_LB_READ_MODE, 0);
	pm2_WR(private, PM2R_LB_SOURCE_OFFSET, 0);
	pm2_WR(private, PM2R_FB_SOURCE_OFFSET, 0);
	pm2_WR(private, PM2R_FB_PIXEL_OFFSET, 0);
	pm2_WR(private, PM2R_WINDOW_ORIGIN, 0);
	pm2_WR(private, PM2R_FB_WINDOW_BASE, 0);
	pm2_WR(private, PM2R_LB_WINDOW_BASE, 0);
	pm2_WR(private, PM2R_FB_SOFT_WRITE_MASK, ~(0L));
	pm2_WR(private, PM2R_FB_HARD_WRITE_MASK, ~(0L));
	pm2_WR(private, PM2R_FB_READ_PIXEL, 0);
	pm2_WR(private, PM2R_DITHER_MODE, 0);
	pm2_WR(private, PM2R_AREA_STIPPLE_MODE, 0);
	pm2_WR(private, PM2R_DEPTH_MODE, 0);
	pm2_WR(private, PM2R_STENCIL_MODE, 0);
	pm2_WR(private, PM2R_TEXTURE_ADDRESS_MODE, 0);
	pm2_WR(private, PM2R_TEXTURE_READ_MODE, 0);
	pm2_WR(private, PM2R_TEXEL_LUT_MODE, 0);
	pm2_WR(private, PM2R_YUV_MODE, 0);
	pm2_WR(private, PM2R_COLOR_DDA_MODE, 0);
	pm2_WR(private, PM2R_TEXTURE_COLOR_MODE, 0);
	pm2_WR(private, PM2R_FOG_MODE, 0);
	pm2_WR(private, PM2R_ALPHA_BLEND_MODE, 0);
	pm2_WR(private, PM2R_LOGICAL_OP_MODE, 0);
	pm2_WR(private, PM2R_STATISTICS_MODE, 0);
	pm2_WR(private, PM2R_SCISSOR_MODE, 0);
	pm2_RDAC_WR(private, PM2I_RD_CURSOR_CONTROL, 0);
	pm2_RDAC_WR(private, PM2I_RD_MISC_CONTROL, PM2F_RD_PALETTE_WIDTH_8);
	pm2_RDAC_WR(private, PM2I_RD_COLOR_KEY_CONTROL, 0);
	pm2_RDAC_WR(private, PM2I_RD_OVERLAY_KEY, 0);
	pm2_RDAC_WR(private, PM2I_RD_RED_KEY, 0);
	pm2_RDAC_WR(private, PM2I_RD_GREEN_KEY, 0);
	pm2_RDAC_WR(private, PM2I_RD_BLUE_KEY, 0);
	if (private->memclock)
		set_memclock(private, private->memclock);
	return 0;
}

static int pm2free_pci_canhandle(pcicfg_vaddr addr) 
{
	kgiu16 vendor,device;

	vendor=pcicfg_in16(addr + PCI_VENDOR_ID);
	device=pcicfg_in16(addr + PCI_DEVICE_ID);

	if (vendor==PCI_VENDOR_ID_TI && 
		device== PCI_DEVICE_ID_TI_TVP4020) {
		NOTICE1("  PCI device is a TI Permedia 2.");
		return 1;
	} else 
	if (vendor==PCI_VENDOR_ID_3DLABS && 
		device== 0x0009) {
		NOTICE1("  PCI device is a 3dlabs Permedia 2v.");
		return 1;
	} 

	/* If more compatible devices turn up, add them here */

	return 0;
}

static int pm2free_verify(struct kgi_display *dpy)
{
	GET_PRIVATE;
	kgi_sint err=0;
	kgiu16  cmd = 0;
	mem_addr base0 = (mem_addr)MEM_NULL;
	mem_addr base1 = KGIM_OPTIONS(dpy, misc)->law_base;
	struct kgim_options_pci *pci = KGIM_OPTIONS(dpy, pci);
	struct kgim_options_misc *misc = KGIM_OPTIONS(dpy, misc);

	DEBUG2("testing for Permedia 2, dev = %.8x", pci->dev);

	memset(private, 0, sizeof(*private));
	private->mem_fb   = MEM_NULL;
	private->mem_regs = MEM_NULL;

	MEM_REGS_REGION   = pm2free_mem_regs;
	MEM_FB_REGION     = pm2free_mem_fb;
	CHIPSET_INFO      = pm2free_info;

	/*	PCI verification is easiest. However, we may have to enable it.
	*/
	if (pci->dev != PCICFG_NULL) {
	
		kgiu16 cmd;
		#define IO_ENABLED      (cmd & PCI_COMMAND_IO)
		#define MEM_ENABLED     (cmd & PCI_COMMAND_MEMORY)
		#define ENABLED	 (cmd & (PCI_COMMAND_IO | \
						PCI_COMMAND_MEMORY))

		if (! pm2free_pci_canhandle(pci->dev)) {
			NOTICE1("  no permedia device detected.");
			return -E(CHIPSET, UNKNOWN);
		}
		
		/*	We check now if the specified LAW window is available,
		**	thus we don't have to restore modifications in case
		**	it is not.
		*/
		cmd   = pcicfg_in16(pci->dev + PCI_COMMAND);
		base0 = pcicfg_in32(pci->dev + PCI_BASE_ADDRESS_0);
		base1 = pcicfg_in32(pci->dev + PCI_BASE_ADDRESS_1);

/* A big endian copy of the registers it there at bas0+PM2_REGS_SIZE 
 */
#ifdef __BIG_ENDIAN
		base0 += PM2_REGS_SIZE;
#endif

/* O.K. - let's get access to the registers.
 */
		MEM_REGS_REGION.base = base0 & PCI_BASE_ADDRESS_MEM_MASK;
		MEM_REGS_REGION.size = PM2_REGS_SIZE;

		if ((MEM_REGS_REGION.base == 0) ||
			mem_check_region(& MEM_REGS_REGION)) {

			ERROR("new mem_regs region busy or unset.");
			return -EBUSY;
		}

		/* Precheck for a 1MB window. We don't know the size yet. */
		MEM_FB_REGION.base   = (misc->law_base 
			? misc->law_base : base1) & PCI_BASE_ADDRESS_MEM_MASK;
		MEM_FB_REGION.size   = 1 MB;

		if ((MEM_FB_REGION.base == 0) ||
			mem_check_region(& MEM_FB_REGION)) {

			ERROR("new mem_fb region looks busy.");
			return -EBUSY;
		}

		MEM_REGS  = mem_claim_region(& MEM_REGS_REGION);

		private->mem_control =mem_in32(MEM_REGS + PM2R_MEM_CONTROL);
		private->boot_address=mem_in32(MEM_REGS + PM2R_BOOT_ADDRESS);
		private->mem_config  =mem_in32(MEM_REGS + PM2R_MEM_CONFIG);
		DEBUG1("defaults: memcont=%x boot_addr=%x memconf=%x.\n",
			private->mem_control,private->boot_address,
			private->mem_config);
		switch (private->mem_config & PM2F_MEM_CONFIG_RAM_MASK) {
			case PM2F_MEM_BANKS_1:
				MEM_FB_REGION.size=0x200000;
				break;
			case PM2F_MEM_BANKS_2:
				MEM_FB_REGION.size=0x400000;
				break;
			case PM2F_MEM_BANKS_3:
				MEM_FB_REGION.size=0x600000;
				break;
			case PM2F_MEM_BANKS_4:
				MEM_FB_REGION.size=0x800000;
				break;
		}
		DEBUG1("Memory detected: %x MB.\n",
			MEM_FB_REGION.size>>20);
		MEM_FB    = mem_claim_region(& MEM_FB_REGION);
		CHIPSET_INFO.ramsize=MEM_FB_REGION.size;
		private->memclock=CVPPC_MEMCLOCK;

		/* Set up the mmio structs */
		MMIO_FB_REGION.next	= NULL;
		MMIO_FB_REGION.name	= "video ram";
		MMIO_FB_REGION.type	= MMAP_TYPE_MMIO|MMAP_PUBLIC | MMAP_FRAMEBUFFER;
		MMIO_FB_REGION.prot	= MMIO_RDWR | MMIO_SHARED;
		MMIO_FB_REGION.psize	= 
		MMIO_FB_REGION.size	= MEM_FB_REGION.size;
		MMIO_FB_REGION.pstart	= MEM_FB_REGION.base;
		MMIO_FB_REGION.offset   = 0;
		MMIO_FB_REGION.set_offset= NULL;
		MMIO_FB_REGION.private	= NULL;
	} else 
		err=1;	/* The PM2 is always PCI AFAIK. */
	
	if (err) {

		if (pci->dev != PCICFG_NULL) {

			pcicfg_out16(cmd, pci->dev + PCI_COMMAND);
		}

		NOTICE2("no Permedia 2 detected");
		return -E(CHIPSET, UNKNOWN);
	}

	DEBUG1("PM2 detected.\n");

	dpy->card->chipset = & CHIPSET_INFO;
	return EOK;
}

int pm2free_find_pci_devices (struct kgi_display *dpy) {
	struct kgim_options_pci *pci = KGIM_OPTIONS(dpy, pci);
	pcicfg_vaddr curr;
	
	if (! pci_present() ) {	
		DEBUG2("No PCI bus found.");
		return 0;
	}

	curr= PCICFG_VADDR(0, 0, 0);
	do {
		if (pm2free_pci_canhandle(curr)) {
			pci->dev=curr;
			return 1;
		}
	} while(pci_get_next_device(&curr));

	return 0;
}

/*
**	KGI module interface
*/
kgi_sint kgim_chipset_init(struct kgi_card *card, struct kgi_display *dpy)
{
	struct chipset_private *private;
	struct kgim_options_pci *pci = KGIM_OPTIONS(dpy, pci);

	/* Inform the user what driver he is using. Very helpful for
	 * bug reports.
	 */
	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	/* Get some private data the driver will need, initialize it, and hook it up.
	 */
	private = kgi_malloc_small(sizeof(*private));
	if (private == NULL) {

		NOTICE2("failed to allocate private data");
		return -ENOMEM;
	}

	memset(private, 0, sizeof(*private));
	KGIM_PRIVATE(dpy, chipset) = private;
	card->chipset = &CHIPSET_INFO;
	
	/* In case the user has not given PCI options, scan for compatible cards 
	 * on the PCI bus.
	 */
	if (pci->dev == PCICFG_NULL) {

		NOTICE2("automatically probing for PCI slot.");

		if (! pm2free_find_pci_devices(dpy)) {

			ERROR("failed to find Permedia 2 on PCI. Give explicit PCI options to override.");
			kgi_free_small(private);
			KGIM_PRIVATE(dpy, chipset) = NULL;
			card->chipset = NULL;
			return -E(CHIPSET, UNKNOWN);
		}
	}

	/* We now have either found a card or the user has specified the PCI
	 * bus/slot/function. We now probe there, if this is really a comaptible
	 * card. If we have the data from the user, it might be wrong, and
	 * there might be incompatible subversions is the manufacturer was
	 * very braindead.
	 */
	if ( pm2free_verify(dpy) ) {

		ERROR("failed to detect Permedia 2.");
		kgi_free_small(private);
		KGIM_PRIVATE(dpy, chipset) = NULL;
		card->chipset = NULL;
		return -E(CHIPSET, UNKNOWN);
	}

	/* O.K. - we have a card, and it is supported by this driver. Reset it.
	 */
	pm2free_reset(dpy);
	return EOK;
}

void kgim_chipset_done(struct kgi_display *dpy)
{
	GET_PRIVATE;
	struct kgim_options_pci *pci = KGIM_OPTIONS(dpy, pci);

	
	if (pci->dev != PCICFG_NULL)
		pcicfg_out16(PCI_COMMAND_IO + PCI_COMMAND_MEMORY,
			pci->dev + PCI_COMMAND);
	
	if (MEM_FB_REGION.baseptr) 
		mem_free_region(&MEM_FB_REGION);
	
	if (MEM_REGS_REGION.baseptr) 
		mem_free_region(&MEM_REGS_REGION);
		
	MEM_REGS = 0;
	MEM_FB = 0;
	
#if 0
#if DEBUG_LEVEL > 1
	DEBUG2("card-state before restore");
	save_state(dpy, &debug_state);
	dump_state(dpy, &debug_state);
#endif
	restore_state(dpy, &reset_state);
	
#if DEBUG_LEVEL > 1
	DEBUG2("card-state after restore");
	save_state(dpy, &debug_state);
	dump_state(dpy, &debug_state);
#endif
#endif
        
	kgi_free_small(private);
	KGIM_PRIVATE(dpy, chipset) = NULL;
	NOTICE(DRIVER_NAME " rev " DRIVER_REV " removed.");
}

static void adjust_timing(struct kgi_mode *mode)
{
	register int foo = (mode->request.graphtype == KGIGT_TEXT16) 
				? mode->request.dpp.x : 8;

	#define ADJUST(X) mode->x.X -= (mode->x.X % foo)

	ADJUST(width);
	ADJUST(blankstart);
	ADJUST(syncstart);
	ADJUST(syncend);
	ADJUST(blankend);

	if (mode->x.total % foo) {

		mode->x.total += foo - (mode->x.total % foo);
	}

	if (((mode->x.syncend / foo) & 0x3F) ==
		((mode->x.syncstart / foo) & 0x3F)) {

		mode->x.syncend -= foo;
	}

	if (((mode->x.blankend / foo) & 0x7F) ==
		((mode->x.blankstart / foo) & 0x7F)) {

		mode->x.blankend -= foo;
	}

	if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F)) {

		mode->y.syncend--;
	}

	if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF)) {

		mode->y.blankend--;
	}

	#undef ADJUST
}

kgi_sint kgim_chipset_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			    enum kgi_tm_cmd *cmd)
{
	GET_PRIVATE;
	int Bpp = 1;
	
	if (mode->flags & FL_MON_INTERLACE) {

		NOTICE1("interlaced modes not supported (yet).\n");
		return -E(CHIPSET, UNKNOWN); 
	}

	switch(mode->request.graphtype) {

	case KGIGT_32BIT:  Bpp =4;break;
	case KGIGT_24BIT:  Bpp =3;break;
	case KGIGT_16BIT:  Bpp =2;break;
	case KGIGT_8BIT:   Bpp =1;break;
	default:
		DEBUG2("graphtype %i not supported", mode->request.graphtype);
		return -E(CHIPSET, UNKNOWN);
	}

	if (mode->lclk.mul != 1) {

		NOTICE1("invalid lclk multiplier");
		return -E(CHIPSET, UNKNOWN);
	}

	switch (mode->lclk.div) {

	case 1:
		break;

	default:
		DEBUG1("invalid CLK to LCLK divider %i", mode->lclk.div);
		return -E(CHIPSET, UNKNOWN);
	}

#define	XVIRTUAL	(mode->request.virt.x)
#define	YVIRTUAL	(mode->request.virt.y)
#define	XVISIBLE	(mode->request.visible.x)
#define	YVISIBLE	(mode->request.visible.y)
#define	XDPP		(mode->request.dpp.x)
#define	YDPP		(mode->request.dpp.y)

	switch (*cmd) {

	case CMD_PROPOSE:

		/* mode->request.graphtype == KGIGT_TEXT16 */

		mode->magnify.x = mode->magnify.y = 1;
			
		if (XVISIBLE == KGI_AUTO)	XVISIBLE = 640;
		if (YVISIBLE == KGI_AUTO)	YVISIBLE = 400;
		if (XVIRTUAL == KGI_AUTO)	XVIRTUAL = XVISIBLE;
		if (YVIRTUAL == KGI_AUTO)	YVIRTUAL = YVISIBLE;
		if (XDPP     == KGI_AUTO)	XDPP     =   1;
		if (YDPP     == KGI_AUTO)	YDPP     =   1;
			
		XVIRTUAL &= 0x7FFFFFF8;
		if (XVIRTUAL > 2048) {

			DEBUG2("%i hpixels are too many", XVIRTUAL);
			return -E(CHIPSET, UNKNOWN);
		}
		if (YVISIBLE < 200) {
			DEBUG2("height not supported: %u\n",
						YVISIBLE);
			return -E(CHIPSET, UNKNOWN);
		}
		if ( YVIRTUAL < 200 || YVIRTUAL > 2048 ) {
			DPRINTK("virtual height not supported: %u\n", 
				YVIRTUAL);
			return -E(CHIPSET, UNKNOWN);
		}
			
		if (YVIRTUAL == 0) {
			YVIRTUAL = CHIPSET_INFO.ramsize / XVIRTUAL;
			DEBUG1("setting yvirtual = %i", YVIRTUAL);
		}
	
		if ((XVIRTUAL * YVIRTUAL * Bpp) >
			CHIPSET_INFO.ramsize) {

			DEBUG2("not enough memory for %ix%i@%ibpp",\
				XVIRTUAL, YVIRTUAL, Bpp * 8);
			return -ENOMEM;
		}

		mode->request.visible.x &= 0x7FFFFFF8;
		mode->request.visible.y &= 0x7FFFFFF8;
			
		mode->x.width = mode->request.visible.x * mode->magnify.x;
		mode->y.width = mode->request.visible.y * mode->magnify.y;
		return EOK;

	case CMD_LOWER:

		if (mode->dclk < CHIPSET_INFO.dclk.min) {

			DEBUG2("%i Hz DCLK are too low", mode->dclk);
			return -E(CHIPSET, UNKNOWN);
		}

		mode->dclk = (mode->dclk > CHIPSET_INFO.dclk.max)
			? CHIPSET_INFO.dclk.max : mode->dclk;

		adjust_timing(mode);
		return EOK;

	case CMD_RAISE:

		if (mode->dclk > CHIPSET_INFO.dclk.max) {

			DEBUG2("%i Hz DCLK are too high", mode->dclk);
			return -E(CHIPSET, UNKNOWN);
		}

		mode->dclk = (mode->dclk < CHIPSET_INFO.dclk.min)
			? CHIPSET_INFO.dclk.min : mode->dclk;

		adjust_timing(mode);
		return EOK;

	case CMD_CHECK:
		switch (XVIRTUAL) {

		case 640:
		case 800:
		case 1024:
		case 1152:
		case 1280:
		case 1600:
		case 2048:
			break;

		default:
			return -E(CHIPSET, UNKNOWN);
		}

		if (((XVIRTUAL * YVIRTUAL * Bpp) > 
			CHIPSET_INFO.ramsize) ||
			(mode->magnify.x != 1)/*!!! temp */ ||
			(mode->magnify.y != 1)/*!!! temp */ ||
			(mode->clk.mul != 1) || (mode->clk.div != 1) ||
			(mode->dclk < CHIPSET_INFO.dclk.min) ||
			(CHIPSET_INFO.dclk.max < mode->dclk)) {
				return -E(CHIPSET, UNKNOWN);
		}

		mode->fb      = MEM_FB;
		mode->fb_size = CHIPSET_INFO.ramsize;
		mode->fsize   = mode->request.virt.x *
			mode->request.virt.y * Bpp;
		mode->io  = NULL;
		mode->mem = NULL;
		mode->mmio = & (private->mmio_fb);

#define TM(X) (((unsigned) CHIPSET_INFO.maxdots.x < mode->x.X)\
		|| ((unsigned) CHIPSET_INFO.maxdots.y<mode->y.X))

	if (TM(width) || TM(blankstart) || TM(syncstart) ||
		TM(syncend) || TM(blankend) || TM(total)) {

		return -E(CHIPSET, UNKNOWN);
	}

#undef TM

	adjust_timing(mode);
	return EOK;

	default:
		INTERNAL_ERROR;
		return -E(CHIPSET, UNKNOWN);
	}

#undef XVIRTUAL 
#undef YVIRTUAL
}

void kgim_chipset_set_mode (struct kgi_display *dpy, struct kgi_mode *mode)
{
	IS_CRITICAL;
	GET_PRIVATE;
	kgiu32 clrmode=0;
	kgiu32 txtmap=0;
	kgiu32 xres;
	kgiu32 video;
	int pixsize;
	int data64;
	int depth;

	DEBUG1("setting mode %i, %ix%i.%ix%i.%ix%i", \
		mode->request.graphtype,\
		mode->request.visible.x, mode->request.visible.y, \
		mode->request.virt.x, mode->request.virt.y, \
		mode->request.dpp.x, mode->request.dpp.y); 

#if 0
#######################
static int pm2fb_decode_var(const struct fb_var_screeninfo* var,
				void* par, struct fb_info_gen* info) {
	struct pm2fb_info* i=(struct pm2fb_info* )info;
	struct pm2fb_par p;
	kgiu32 xres;

	memset(&p, 0, sizeof(struct pm2fb_par));
	p.width=(var->xres_virtual+7)&~7;
	p.height=var->yres_virtual;
/*	p.depth=(var->bits_per_pixel+7)&~7; */
/*	p.depth=p.depth>32?32:p.depth; */
/*	data64=p.depth>8; */
	xres=(var->xres+31)&~31;
	if (p.width<xres+var->xoffset)
		p.width=xres+var->xoffset;
	if (p.height<var->yres+var->yoffset)
		p.height=var->yres+var->yoffset;
	if (!partprod(xres)) {
		DPRINTK("width not supported: %u\n", xres);
		return -EINVAL;
	}
	if (p.width*p.height*p.depth/8>i->regions.fb_size) {
		DPRINTK("no memory for screen (%ux%ux%u)\n",
						p.width, p.height, p.depth);
		return -EINVAL;
	}


	*((struct pm2fb_par* )par)=p;
	return 0;
}
#endif

	xres=(mode->request.virt.x+31)&~31;
	set_aperture(private, mode->request.graphtype);

	WAIT_FIFO(private, 22);
	pm2_RDAC_WR(private, PM2I_RD_COLOR_KEY_CONTROL, 
			mode->request.graphtype==KGIGT_8BIT ? 0:
						PM2F_COLOR_KEY_TEST_OFF);
	switch (mode->request.graphtype) {
		case KGIGT_8BIT:
			pm2_WR(private, PM2R_FB_READ_PIXEL, 0);
			pixsize=1;
			depth=8;
			break;
		case KGIGT_16BIT:
			pm2_WR(private, PM2R_FB_READ_PIXEL, 1);
			clrmode=PM2F_RD_TRUECOLOR|0x06;
			txtmap=PM2F_TEXTEL_SIZE_16;
			pixsize=2;
			depth=16;
			break;
		case KGIGT_32BIT:
			pm2_WR(private, PM2R_FB_READ_PIXEL, 2);
			clrmode=PM2F_RD_TRUECOLOR|0x08;
			txtmap=PM2F_TEXTEL_SIZE_32;
			pixsize=4;
			depth=32;
			break;
		case KGIGT_24BIT:
			pm2_WR(private, PM2R_FB_READ_PIXEL, 4);
			clrmode=PM2F_RD_TRUECOLOR|0x09;
			txtmap=PM2F_TEXTEL_SIZE_24;
			pixsize=3;
			depth=24;
			break;
		default:
			return;	/* WHAAAT ? We should never have gotten here ... */
	}
	data64=pixsize>1;
	DEBUG3("data64=%d",data64);

	pm2_WR(private, PM2R_SCREEN_SIZE, (mode->request.virt.y<<16)|mode->request.virt.x);
	pm2_WR(private, PM2R_SCISSOR_MODE, PM2F_SCREEN_SCISSOR_ENABLE);
	pm2_WR(private, PM2R_FB_WRITE_MODE, PM2F_FB_WRITE_ENABLE);
	pm2_WR(private, PM2R_FB_READ_MODE, partprod(xres));
	pm2_WR(private, PM2R_LB_READ_MODE, partprod(xres));
	pm2_WR(private, PM2R_TEXTURE_MAP_FORMAT, txtmap|partprod(xres));
	pm2_WR(private, PM2R_H_TOTAL, to3264(mode->x.total, depth, data64)-1);
	pm2_WR(private, PM2R_HS_START, to3264(mode->x.syncstart-mode->x.width, depth, data64));
#if 0
	p.vsstart=var->lower_margin?var->lower_margin-1:0;	/* FIXME! */
	p.vsend=var->lower_margin+var->vsync_len-1;
	p.vbend=var->lower_margin+var->vsync_len+var->upper_margin;
	p.vtotal=var->yres+p.vbend-1;
/*	p.base=to3264(var->yoffset*xres+var->xoffset, p.depth, 1); */
#endif
	pm2_WR(private, PM2R_HS_END, 	to3264(mode->x.syncend -mode->x.width, depth, data64));
	pm2_WR(private, PM2R_HG_END, 	to3264(mode->x.total   -mode->x.width, depth, data64));
	pm2_WR(private, PM2R_HB_END, 	to3264(mode->x.total   -mode->x.width, depth, data64));

/* One might expect here:
 *	pm2_WR(private, PM2R_HB_END, 	to3264(mode->x.blankend-mode->x.width, depth, data64));
 * But that doesn't work. You will get garbage on the left border, if you do this.
 */
	pm2_WR(private, PM2R_V_TOTAL, mode->y.total);
	pm2_WR(private, PM2R_VS_START,  mode->y.syncstart-mode->y.width);
	pm2_WR(private, PM2R_VS_END, 	mode->y.syncend  -mode->y.width);
/* One might expect here:
	pm2_WR(private, PM2R_VB_END,    mode->y.blankend -mode->y.width);
 * But that doesn't work. You will get garbage on the lower border, if you do this.
 */
	pm2_WR(private, PM2R_VB_END,    mode->y.total    -mode->y.width+1);
	pm2_WR(private, PM2R_SCREEN_STRIDE, to3264(mode->request.virt.x, depth, 1));
	DEFW();
	pm2_WR(private, PM2R_SCREEN_BASE, 0);	/* Panning - start at 0 */
	pm2_RDAC_WR(private, PM2I_RD_COLOR_MODE, PM2F_RD_COLOR_MODE_RGB|
						PM2F_RD_GUI_ACTIVE|clrmode);

	video=0;
	if (data64)
		video|=PM2F_DATA_64_ENABLE;
	video|= (mode->x.polarity > 0) ? 
			PM2F_HSYNC_ACT_HIGH :
			PM2F_HSYNC_ACT_LOW  ;
	video|= (mode->y.polarity > 0) ? 
			PM2F_VSYNC_ACT_HIGH :
			PM2F_VSYNC_ACT_LOW  ;
	video|=PM2F_VIDEO_ENABLE;

#if 0
	if ((var->vmode & FB_VMODE_MASK)==FB_VMODE_DOUBLE)
		p.video|=PM2F_LINE_DOUBLE;
	if (var->activate==FB_ACTIVATE_NOW)
		p.video|=PM2F_VIDEO_ENABLE;
#endif
	pm2_WR(private, PM2R_VIDEO_CONTROL, video);

	ENTER_CRITICAL;

	kgim_monitor_set_mode(dpy, mode);
	NOTICE2("monitor set");
	kgim_ramdac_set_mode(dpy, mode);
	NOTICE2("ramdac set");
	kgim_clock_set_mode(dpy, mode);
	NOTICE2("clock set");

	LEAVE_CRITICAL;

	set_pixclock(private, mode->dclk/1000); /* Hz vs. kHz p->pixclock); */

	memcpy(&(dpy->mode), mode, sizeof(dpy->mode));
	NOTICE2("setting up display...");

	private->split = -1;
	private->org_xdot = private->org_ydot = private->org_offs = -1;

	dpy->fb.size.x	= mode->request.virt.x;
	dpy->fb.size.y	= mode->request.virt.y;
	dpy->fb.pbuf	= NULL; /* MEM_FB.baseptr; */
	dpy->fb.zbuf	= NULL;
	dpy->fb.ops	= NULL;

	private->fb_size = dpy->fb.size.x * dpy->fb.size.y;
	
	dpy->show_pointer	= NULL;
	dpy->hide_pointer	= NULL;
	dpy->undo_pointer	= NULL;

	dpy->show_cursor	= NULL;
	dpy->hide_cursor	= NULL;
	dpy->undo_cursor	= NULL;

	dpy->set_font		= NULL;
	dpy->set_split		= NULL; /* Can the PM do splitline ? */
	dpy->set_frame		= pm2free_set_frame;
	dpy->set_origin 	= pm2free_set_origin;

#if 0	/* FIXME ! - use the virtual size for that. */
	MMIO_FB_REGION.psize	= 
	MMIO_FB_REGION.size	= MEM_FB_REGION.size;
#endif

	kgim_accel_set_mode(dpy, mode);
	NOTICE2("graphics set");

	NOTICE2("mode set.");
	
	DEBUG1("%.3i MHz %.3i kHz %.3i Hz", \
		mode->dclk / 1000000, \
		(mode->dclk / ((mode->x.total ) - 5)) / 1000, \
		(mode->dclk / ((mode->x.total ) - 5)) / (mode->y.total - 2) );
}

static kgi_sint chipset_command(struct kgi_graphic *state,kgi_uint cmd,kgi_uint arg)
{
	kgi_coord  *coord;
	struct kgi_display *dpy;

	dpy = state->dev.dpy;
	switch (cmd) {
		case CHIP_SETSPLITLINE:
			if (arg > state->dev.mode.virt.y)
				return -E(CHIPSET,BOGUS);

			kgi_set_split(&(state->dev),arg*dpy->mode.magnify.y);
			return 0;
		case CHIP_SETVISFRAME:
			coord = (kgi_coord *)(state->io_buf);

			if (coord->x < 0 || 
			    coord->x > state->dev.mode.virt.x)
				return -E(CHIPSET,BOGUS);

			if (coord->y < 0 || 
			    coord->y > state->dev.mode.virt.y)
				return -E(CHIPSET,BOGUS);

			kgi_set_origin(&(state->dev), 0, coord->x, coord->y);
			return 0;
		default:
			break;
	}

	return -E(CHIPSET,UNKNOWN);
}

kgi_sint kgim_chipset_command(struct kgi_graphic *graph, kgi_uint cmd, kgi_uint arg)
{
	return chipset_command(graph, cmd, arg);
}

